"""Tyler - A framework for AI agents with a complete lack of conventional limitations"""

__version__ = "0.6.0"

from tyler.utils.logging import get_logger

# Configure logging when package is imported
logger = get_logger(__name__) 